@@ -26,40 +26,14 @@
 # WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 
 import sys, os, codecs
-from urllib.parse import unquote
-from urllib.parse import urlsplit
-
-ASCII_CHARS   = set(chr(x) for x in range(128))
-URL_SAFE      = set('ABCDEFGHIJKLMNOPQRSTUVWXYZ'
-                    'abcdefghijklmnopqrstuvwxyz'
-                    '0123456789' '#' '_.-/~')
-IRI_UNSAFE = ASCII_CHARS - URL_SAFE
-
-def quoteurl(href):
-    if isinstance(href,bytes):
-        href = href.decode('utf-8')
-    (scheme, netloc, path, query, fragment) = urlsplit(href, scheme="", allow_fragments=True)
-    if scheme != "":
-        scheme += "://"
-        href = href[len(scheme):]
-    result = []
-    for char in href:
-        if char in IRI_UNSAFE:
-            char = "%%%02x" % ord(char)
-        result.append(char)
-    return scheme + ''.join(result)
-
-def unquoteurl(href):
-    if isinstance(href,bytes):
-        href = href.decode('utf-8')
-    href = unquote(href)
-    return href
+from collections import OrderedDict
+from hrefutils import urlencodepart, urldecodepart
 
 # encode to make xml safe
 def xmlencode(data):
     if data is None:
         return ''
-    newdata = data
+    newdata = xmldecode(data)
     newdata = newdata.replace('&', '&amp;')
     newdata = newdata.replace('<', '&lt;')
     newdata = newdata.replace('>', '&gt;')
@@ -77,11 +51,11 @@ def xmldecode(data):
     newdata = newdata.replace('&amp;', '&')
     return newdata
 
-SPECIAL_HANDLING_TAGS = {
-    '?xml'     : ('xmlheader', -1), 
-    '!--'      : ('comment', -3),
-    '!DOCTYPE' : ('doctype', -1),
-}
+SPECIAL_HANDLING_TAGS = OrderedDict([
+    ('?xml', ('xmlheader', -1)),
+    ('!--',  ('comment', -3)),
+    ('!DOCTYPE', ('doctype', -1))
+    ])
 
 SPECIAL_HANDLING_TYPES = ['xmlheader', 'doctype', 'comment']
 
@@ -100,8 +74,10 @@ class Opf_Parser(object):
         self.spine=[]
         self.guide=[]
         self.bindings=[]
+        self.ns_remap = False
         self._parseData()
 
+
     # OPF tag iterator
     def _opf_tag_iter(self):
         tcontent = last_tattr = None
@@ -114,6 +90,10 @@ class Opf_Parser(object):
                 tcontent = text.rstrip(" \r\n")
             else: # we have a tag
                 ttype, tname, tattr = self._parsetag(tag)
+                # remap opf namespace on tags if needed
+                if tname.startswith('opf:'):
+                    self.ns_remap = True
+                    tname = tname[4:]
                 if ttype == "begin":
                     tcontent = None
                     prefix.append(tname)
@@ -126,7 +106,7 @@ class Opf_Parser(object):
                         prefix.pop()
                         tattr = last_tattr
                         if tattr is None:
-                            tattr = {}
+                            tattr = OrderedDict()
                         last_tattr = None
                     elif ttype == 'single':
                         tcontent = None
@@ -142,10 +122,17 @@ class Opf_Parser(object):
             if tname == "package":
                 ver = tattr.pop("version", "2.0")
                 uid = tattr.pop("unique-identifier","bookid")
+                if self.ns_remap:
+                    if "xmlns:opf" in tattr:
+                        tattr.pop("xmlns:opf")
+                        tattr["xmlns"] = "http://www.idpf.org/2007/opf"
                 self.package = (ver, uid, tattr)
                 continue
             # metadata
             if tname == "metadata":
+                if self.ns_remap:
+                    if not "xmlns:opf" in tattr:
+                        tattr["xmlns:opf"] = "http://www.idpf.org/2007/opf"
                 self.metadata_attr = tattr
                 continue
             if tname in ["meta", "link"] or tname.startswith("dc:") and "metadata" in prefix:
@@ -156,9 +143,13 @@ class Opf_Parser(object):
                 nid = "xid%03d" %  cnt
                 cnt += 1
                 id = tattr.pop("id",nid)
+                # must keep all hrefs in encoded) form
+                # if relative, then no fragments so decode and then encode for safety
                 href = tattr.pop("href","")
+                if href.find(':') == -1:
+                    href = urldecodepart(href)
+                    href = urlencodepart(href)
                 mtype = tattr.pop("media-type","")
-                href = unquoteurl(href)
                 self.manifest.append((id, href, mtype, tattr))
                 continue
             # spine
@@ -173,7 +164,8 @@ class Opf_Parser(object):
             if tname == "reference" and  "guide" in prefix:
                 type = tattr.pop("type","")
                 title = tattr.pop("title","")
-                href = unquoteurl(tattr.pop("href",""))
+                # must keep all hrefs in quoted (encoded) form
+                href = tattr.pop("href","")
                 self.guide.append((type, title, href))
                 continue
             # bindings
@@ -216,15 +208,15 @@ class Opf_Parser(object):
         p = 1
         tname = None
         ttype = None
-        tattr = {}
+        tattr = OrderedDict()
         while p < n and s[p:p+1] == ' ' : p += 1
         if s[p:p+1] == '/':
             ttype = 'end'
             p += 1
             while p < n and s[p:p+1] == ' ' : p += 1
         b = p
-        # handle comment special case as there may be no spaces to 
-        # delimit name begin or end 
+        # handle comment special case as there may be no spaces to
+        # delimit name begin or end
         if s[b:].startswith('!--'):
             p = b+3
             tname = '!--'
@@ -234,8 +226,8 @@ class Opf_Parser(object):
         while p < n and s[p:p+1] not in ('>', '/', ' ', '"', "'","\r","\n") : p += 1
         tname=s[b:p].lower()
         # remove redundant opf: namespace prefixes on opf tags
-        if tname.startswith("opf:"):
-            tname = tname[4:]
+        # if tname.startswith("opf:"):
+        #    tname = tname[4:]
         # more special cases
         if tname == '!doctype':
             tname = '!DOCTYPE'
@@ -256,7 +248,8 @@ class Opf_Parser(object):
                     qt = s[p:p+1]
                     p = p + 1
                     b = p
-                    while p < n and s[p:p+1] != qt: p += 1
+                    # try to work around missing end quotes
+                    while p < n and s[p:p+1] not in ['>', '<', qt] : p += 1
                     val = s[b:p]
                     p += 1
                 else :
@@ -318,7 +311,7 @@ class Opf_Parser(object):
         for (gtype, gtitle,  ghref) in self.guide:
             guide.append((gtype, gtitle, ghref))
         return guide
-    
+
     def get_bindings(self):
         bindings = []
         for (mtype, handler) in self.bindings:
@@ -335,7 +328,7 @@ class Opf_Parser(object):
             xmlres.append(' %s="%s"' % (key, val))
         xmlres.append('>\n')
         return "".join(xmlres)
-                          
+
     def convert_metadata_attr_to_xml(self):
         xmlres = []
         attr = self.metadata_attr
@@ -346,7 +339,7 @@ class Opf_Parser(object):
             xmlres.append(' %s="%s"' % (key, val))
         xmlres.append('>\n')
         return "".join(xmlres)
-        
+
     def convert_metadata_entries_to_xml(self):
         xmlres = []
         for (mname, mcontent, attr) in self.metadata:
@@ -365,8 +358,8 @@ class Opf_Parser(object):
     def convert_manifest_entries_to_xml(self):
         xmlres = []
         for (id, href, mtype, attr) in self.manifest:
-            url = quoteurl(href)
-            xmlres.append('    <item id="%s" href="%s" media-type="%s"' % (id, url, mtype))
+            # all hrefs should be kept in quoted (encoded) form
+            xmlres.append('    <item id="%s" href="%s" media-type="%s"' % (id, href, mtype))
             for key in attr:
                 val= attr[key]
                 val= xmlencode(val)
@@ -378,10 +371,11 @@ class Opf_Parser(object):
         xmlres = []
         attr = self.spine_attr
         xmlres.append('  <spine')
-        for key in attr:
-            val= attr[key]
-            val= xmlencode(val)
-            xmlres.append(' %s="%s"' % (key, val))
+        if attr is not None:
+            for key in attr:
+                val= attr[key]
+                val= xmlencode(val)
+                xmlres.append(' %s="%s"' % (key, val))
         xmlres.append('>\n')
         return "".join(xmlres)
 
@@ -389,20 +383,21 @@ class Opf_Parser(object):
         xmlres=[]
         for (idref, attr) in self.spine:
             xmlres.append('    <itemref idref="%s"' % idref)
-            for key in attr:
-                val= attr[key]
-                val= xmlencode(val)
-                xmlres.append(' %s="%s"' % (key, val))
+            if attr is not None:
+                for key in attr:
+                    val= attr[key]
+                    val= xmlencode(val)
+                    xmlres.append(' %s="%s"' % (key, val))
             xmlres.append('/>\n')
         return "".join(xmlres)
-    
+
     def convert_guide_entries_to_xml(self):
         xmlres=[]
         for (gtype, gtitle, ghref) in self.guide:
-            url = quoteurl(ghref)
-            xmlres.append('    <reference type="%s" title="%s" href="%s"/>\n' % (gtype, gtitle, url))
+            # all hrefs should already be in quoted (encoded) form
+            xmlres.append('    <reference type="%s" title="%s" href="%s"/>\n' % (gtype, gtitle, ghref))
         return "".join(xmlres)
-                          
+
     def convert_binding_entries_to_xml(self):
         xmlres=[]
         for (mtype, handler) in self.bindings:
@@ -423,13 +418,13 @@ class Opf_Parser(object):
         xmlres.append(self.convert_spine_entries_to_xml())
         xmlres.append('  </spine>\n')
         (opfver, uid, attr) = self.package
-        if len(self.guide) > 0 or opfver.startswith('2'):
+        if len(self.guide) > 0:
             xmlres.append('  <guide>\n')
             xmlres.append(self.convert_guide_entries_to_xml())
             xmlres.append('  </guide>\n')
-        if len(self.bindings) > 0 and opver.startswith('3'):
+        if len(self.bindings) > 0 and opfver.startswith('3'):
             xmlres.append('  <bindings>\n')
-            xmlres.append(self.guide.convert_binding_entries_to_xml())
+            xmlres.append(self.convert_binding_entries_to_xml())
             xmlres.append('  </bindings>\n')
         xmlres.append('</package>\n')
         return "".join(xmlres)
@@ -462,10 +457,7 @@ def main():
     # print(op.get_guide())
     # print(op.get_bindings())
     print(op.rebuild_opfxml())
-    return 0    
-
-
-
+    return 0
 
 if __name__ == '__main__':
     sys.exit(main())
